using System;
using System.IO;
using System.Collections;
using System.Collections.Generic;
using System.Text;

namespace Tsukikage.BMSPlayer
{
    namespace Sequence
    {
        public class BmsEventCode
        {
            public const int TypeSeqCtrl = 0x010000;

            public const int PlayBackChorus = 0x010001;
            public const int _ChangeBarLength = 0x010002;
            public const int ChangeBPM = 0x010003;
            public const int ChangeBMP = 0x010004;
            public const int _ExtChr = 0x010005;
            public const int ChangeBMPPoor = 0x010006;
            public const int ChangeBMPLayer = 0x010007;
            public const int ChangeBPMFromList = 0x010008;
            public const int StopSequenceFromList = 0x010009;

            public const int ChangeTextFromList = 0x010099;

            public const int StopSequence = 0x0100F0;
            public const int BarLine = 0x0100F1;

            public const int TypeNote = 0x020000;
            public const int NoteObj = 0x020100;
            public const int NoteOn = 0x020200;
            public const int NoteOff = 0x020300;
            public const int NoteHid = 0x020400;


            public const int TypeJudge = 0x030000;
            public const int JudgeWaveAssign = 0x030100;
            public const int JudgePoorNote = 0x030200;
            public const int JudgePoorNoteOn = 0x030300;
            public const int JudgePoorNoteOff = 0x030400;
            public const int JudgeStartNote = 0x030500;
            public const int JudgeStartNoteOn = 0x030600;
            public const int JudgeStartNoteOff = 0x030700;

            public const int TypeRecord = 0x040000;

        }

        public struct BmsEvent
        {
            public int TickCount;
            public int Event;
            public int Data;
            public int Data2;

            public void Set(int Tick, int Event, int Data)
            {
                this.TickCount = Tick;
                this.Event = Event;
                this.Data = Data;
                this.Data2 = 0;
            }

            public void Set(int Tick, int Event, int Data, int Data2)
            {
                this.TickCount = Tick;
                this.Event = Event;
                this.Data = Data;
                this.Data2 = Data2;
            }

            public int EventType
            {
                get
                {
                    return this.Event & 0xFF0000;
                }
            }

            public int EventAttr
            {
                get
                {
                    return this.Event & 0xFFFF00;
                }
            }

            public int EventLane
            {
                get
                {
                    return this.Event & 0xFF;
                }
            }

            public override string ToString()
            {
                return "t" + TickCount + ":e" + Event + ":d" + Data + "";
            }
        }

        public struct BmsBarInfo
        {
            public int StartTick;
            public int LengthTick;
            public double LengthRate;
        }

        public class BmsInfo
        {
            public string title;
            public string artist;
            public string genre;
            public string stagefile;

            public int player;
            public int rank;
            public int playlevel;
            public int volwav;
            public int total;
            public double bpm;

            public string url;
            public string mail;

            public bool useLongNote;
            public bool is7Key;
            public bool hasBga;

            public int totalnotes;
            public int playtime; // ms

            public BmsInfo()
            {
                title = "";
                artist = "";
                genre = "";
                stagefile = "";

                player = 1;
                rank = 3;
                playlevel = 2;
                volwav = 100;
                total = 1000;
                bpm = 130.0;
                url = "";
                mail = "";

                useLongNote = false;
                is7Key = false;

                totalnotes = 0;
                playtime = 0;

            }

            public override string ToString()
            {
                return "(x" + playlevel + ") [" + genre + "] " + title + " - " + artist;
            }
        }

        public class Bms
        {
            private const int ValueOfZZ = 1296;
            private const int MaxEvents = 65536;

            public BmsInfo infoData = new BmsInfo();
            public BmsEvent[] eventData = new BmsEvent[MaxEvents];
            public BmsBarInfo[] barInfo = new BmsBarInfo[1000];

            public int events = 0;

            public List<string> bmsText = new List<string>();
            public int resolution = 192;
            public int rndSeed = 0;

            public int LNOBJ = 0;

            public string[] wavlist = new string[ValueOfZZ];
            public string[] bmplist = new string[ValueOfZZ];
            public string[] textlist = new string[ValueOfZZ];
            public double[] bpmlist = new double[ValueOfZZ];
            public int[] stoplist = new int[ValueOfZZ];


            public Bms(Stream source)
            {
                if (source == null)
                    throw new ArgumentNullException("source");

                using (StreamReader sr = new StreamReader(source, Encoding.Default))
                {
                    string s;
                    while ((s = sr.ReadLine()) != null)
                        bmsText.Add(s);

                    sr.Close();
                }

                //bmsText = (string[])ar.ToArray();
                rndSeed = new Random().Next();
            }

            public string GetStageFileName()
            {
                foreach (string linestr in bmsText)
                {
                    string uppered = linestr.ToUpper();

                    if (uppered.IndexOf("#STAGEFILE ") == 0)
                        return linestr.Substring("#STAGEFILE ".Length);
                }

                return "";
            }

            public void ParseInfo()
            {
                foreach (string linestr in bmsText)
                {
                    string uppered = linestr.ToUpper();
                    if (uppered.IndexOf("#TITLE ") == 0) { infoData.title = linestr.Substring("#TITLE ".Length); continue; }
                    if (uppered.IndexOf("#ARTIST ") == 0) { infoData.artist = linestr.Substring("#ARTIST ".Length); continue; }
                    if (uppered.IndexOf("#GENRE ") == 0) { infoData.genre = linestr.Substring("#GENRE ".Length); continue; }
                    if (uppered.IndexOf("#STAGEFILE ") == 0) { infoData.stagefile = linestr.Substring("#STAGEFILE ".Length); continue; }

                    if (uppered.IndexOf("#PLAYER ") == 0) { infoData.player = ParseInt(linestr.Substring("#PLAYER ".Length)); continue; }
                    if (uppered.IndexOf("#RANK ") == 0) { infoData.rank = ParseInt(linestr.Substring("#RANK ".Length)); continue; }
                    if (uppered.IndexOf("#PLAYLEVEL ") == 0) { infoData.playlevel = ParseInt(linestr.Substring("#PLAYLEVEL ".Length)); continue; }
                    if (uppered.IndexOf("#VOLWAV ") == 0) { infoData.volwav = ParseInt(linestr.Substring("#VOLWAV ".Length)); continue; }
                    if (uppered.IndexOf("#TOTAL ") == 0) { infoData.total = ParseInt(linestr.Substring("#TOTAL ".Length)); continue; }
                    if (uppered.IndexOf("#BPM ") == 0) { infoData.bpm = ParseDouble(linestr.Substring("#BPM ".Length)); continue; }

                    if (uppered.IndexOf("%URL ") == 0) { infoData.url = linestr.Substring("%URL ".Length); continue; }
                    if (uppered.IndexOf("%MAIL ") == 0) { infoData.mail = linestr.Substring("%MAIL ".Length); continue; }

                    if (uppered.IndexOf("#LNOBJ ") == 0) { LNOBJ = ParseInt36(linestr.Substring("#LNOBJ ".Length)); continue; }

                    if (uppered.IndexOf("#WAV") == 0)
                    {
                        int id = ParseInt36(linestr.Substring("#WAV".Length, 2));
                        if (id < 0 || id > 1295) continue;
                        wavlist[id] = linestr.Substring("#WAV__ ".Length);
                    }
                }
            }

            public void Parse()
            {
                eventData.Initialize();

                int barCnt = 0; // ߐ

                int rndCurValue = -1;
                int[] rndValue = new int[1000];

                Random rnd = new Random(rndSeed);
                int rndCnt = 0;

                bool ignore = false;
                foreach (string linestr in bmsText)
                {
                    string uppered = linestr.ToUpper();

                    if (uppered.IndexOf("#ENDIF") == 0) { ignore = false; continue; }
                    if (ignore) continue;
                    if (uppered.IndexOf("#RANDOM ") == 0)
                    {
                        int rndMax = ParseInt(linestr.Substring("#RANDOM ".Length));
                        if (rndMax == 0) continue;

                        rndCurValue = rndValue[rndCnt++] = rnd.Next(rndMax);
                        continue;
                    }

                    if (uppered.IndexOf("#IF ") == 0)
                    {
                        if (rndCurValue != ParseInt(linestr.Substring("#IF ".Length)))
                            ignore = true;

                        continue;
                    }

                    if (uppered.IndexOf("#TITLE ") == 0) { infoData.title = linestr.Substring("#TITLE ".Length); continue; }
                    if (uppered.IndexOf("#ARTIST ") == 0) { infoData.artist = linestr.Substring("#ARTIST ".Length); continue; }
                    if (uppered.IndexOf("#GENRE ") == 0) { infoData.genre = linestr.Substring("#GENRE ".Length); continue; }
                    if (uppered.IndexOf("#STAGEFILE ") == 0) { infoData.stagefile = linestr.Substring("#STAGEFILE ".Length); continue; }

                    if (uppered.IndexOf("#PLAYER ") == 0) { infoData.player = ParseInt(linestr.Substring("#PLAYER ".Length)); continue; }
                    if (uppered.IndexOf("#RANK ") == 0) { infoData.rank = ParseInt(linestr.Substring("#RANK ".Length)); continue; }
                    if (uppered.IndexOf("#PLAYLEVEL ") == 0) { infoData.playlevel = ParseInt(linestr.Substring("#PLAYLEVEL ".Length)); continue; }
                    if (uppered.IndexOf("#VOLWAV ") == 0) { infoData.volwav = ParseInt(linestr.Substring("#VOLWAV ".Length)); continue; }
                    if (uppered.IndexOf("#TOTAL ") == 0) { infoData.total = ParseInt(linestr.Substring("#TOTAL ".Length)); continue; }
                    if (uppered.IndexOf("#BPM ") == 0) { infoData.bpm = ParseDouble(linestr.Substring("#BPM ".Length)); continue; }

                    if (uppered.IndexOf("%URL ") == 0) { infoData.url = linestr.Substring("%URL ".Length); continue; }
                    if (uppered.IndexOf("%MAIL ") == 0) { infoData.mail = linestr.Substring("%MAIL ".Length); continue; }

                    if (uppered.IndexOf("#LNOBJ ") == 0) { LNOBJ = ParseInt36(linestr.Substring("#LNOBJ ".Length)); continue; }


                    if (uppered.IndexOf("#WAV") == 0)
                    {
                        int id = ParseInt36(linestr.Substring("#WAV".Length, 2));
                        if (id < 0 || id > 1295) continue;
                        wavlist[id] = linestr.Substring("#WAV__ ".Length);
                    }

                    if (uppered.IndexOf("#BMP") == 0)
                    {
                        int id = ParseInt36(linestr.Substring("#BMP".Length, 2));
                        if (id < 0 || id > 1295) continue;
                        bmplist[id] = linestr.Substring("#BMP__ ".Length);
                    }

                    if (uppered.IndexOf("#TEXT") == 0)
                    {
                        int id = ParseInt36(linestr.Substring("#TEXT".Length, 2));
                        if (id < 0 || id > 1295) continue;
                        textlist[id] = linestr.Substring("#TEXT__ ".Length);
                    }

                    if (uppered.IndexOf("#BPM") == 0)
                    {
                        int id = ParseInt36(linestr.Substring("#BPM".Length, 2));
                        if (id < 0 || id > 1295) continue;
                        double value = ParseDouble(linestr.Substring("#BPM__ ".Length));
                        bpmlist[id] = value;
                    }

                    if (uppered.IndexOf("#STOP") == 0)
                    {
                        int id = ParseInt36(linestr.Substring("#STOP".Length, 2));
                        if (id < 0 || id > 1295) continue;
                        stoplist[id] = ParseInt(linestr.Substring("#STOP__ ".Length));
                    }

                    if ((uppered.Length > 7) && (uppered[0] == '#') && ((uppered[1] & 0xF0) == 0x30) && ((uppered[2] & 0xF0) == 0x30) && ((uppered[3] & 0xF0) == 0x30))
                    {
                        int br = ParseInt(linestr.Substring(1, 3));
                        int ch = ParseInt16(linestr.Substring(4, 2));
                        if (ch == -1) continue;
                        string dt = linestr.Substring("#XXXYY ".Length);

                        if (barCnt <= br)
                            barCnt = br + 1;

                        if (ch == 0x02)
                        {
                            barInfo[br].LengthRate = ParseDouble(dt);
                        }
                    }
                }

                {
                    int tick = 0;
                    for (int br = 0; br < barCnt; br++)
                    {
                        if (barInfo[br].LengthRate == 0.0f)
                            barInfo[br].LengthRate = 1.0f;
                        barInfo[br].StartTick = tick;

                        AddEvent(tick, BmsEventCode.BarLine, br);

                        tick += (int)(barInfo[br].LengthRate * resolution);
                    }
                }

                rndCnt = 0;

                ignore = false;
                int[] LastNoteObjEventID = new int[32];
                int[] LastNoteOnEventID = new int[32];

                foreach (string linestr in bmsText)
                {
                    string uppered = linestr.ToUpper();

                    if (uppered.IndexOf("#ENDIF") == 0) { ignore = false; continue; }
                    if (ignore) continue;
                    if (uppered.IndexOf("#RANDOM ") == 0)
                    {
                        rndCurValue = rndValue[rndCnt++];
                        continue;
                    }
                    if (uppered.IndexOf("#IF ") == 0)
                    {
                        if (rndCurValue != ParseInt(linestr.Substring("#IF ".Length)))
                            ignore = true;
                        continue;
                    }

                    if ((uppered.Length > 7) && (uppered[0] == '#') && ((uppered[1] & 0xF0) == 0x30) && ((uppered[2] & 0xF0) == 0x30) && ((uppered[3] & 0xF0) == 0x30))
                    {
                        int br = ParseInt(linestr.Substring(1, 3));
                        int ch = ParseInt16(linestr.Substring(4, 2));
                        if (ch == -1) continue;
                        if (ch == 0x02) continue;

                        string dtstr = linestr.Substring("#XXXYY ".Length);

                        int dtcnt = 0;
                        int dtlen = dtstr.Length / 2;
                        while (dtcnt < dtlen && ParseInt36(dtstr.Substring(dtcnt++ * 2, 2)) != -1) ;

                        for (int i = 0; i < dtcnt; i++)
                        {
                            int dt = ParseInt36(dtstr.Substring(i * 2, 2));
                            if (dt == -1) break;
                            if (dt == 0) continue;

                            int tick = (int)(barInfo[br].StartTick + barInfo[br].LengthRate * resolution * i / dtcnt);

                            if (ch < 0x10) // ch:0x
                            {
                                switch (ch)
                                {
                                    case 0x01: AddEvent(tick, BmsEventCode.PlayBackChorus, dt); break;
                                    case 0x02: break;
                                    case 0x03: AddEvent(tick, BmsEventCode.ChangeBPM, dt / 36 * 16 + dt % 36); break;
                                    case 0x04: AddEvent(tick, BmsEventCode.ChangeBMP, dt); infoData.hasBga = true; ; break;
                                    case 0x05: break;
                                    case 0x06: AddEvent(tick, BmsEventCode.ChangeBMPPoor, dt); infoData.hasBga = true; break;
                                    case 0x07: AddEvent(tick, BmsEventCode.ChangeBMPLayer, dt); infoData.hasBga = true; break;
                                    case 0x08: AddEvent(tick, BmsEventCode.ChangeBPMFromList, dt); break;
                                    case 0x09: AddEvent(tick, BmsEventCode.StopSequenceFromList, dt); break;
                                }
                            }

                            else if (ch < 0x30) // ch: 1x, 2x 
                            {
                                int Lane = ch - 0x10;
                                if (Lane == 0x08 || Lane == 0x09 || Lane == 0x18 || Lane == 0x19)
                                    infoData.is7Key = true;

                                if (dt == LNOBJ && LastNoteObjEventID[Lane] != 0)
                                {
                                    eventData[LastNoteObjEventID[Lane]].Event = BmsEventCode.NoteOn + Lane;
                                    eventData[LastNoteObjEventID[Lane]].Data2 = tick;
                                    AddEvent(tick, BmsEventCode.NoteOff + Lane, eventData[LastNoteObjEventID[Lane]].Data);
                                    LastNoteObjEventID[Lane] = 0;
                                }
                                else
                                {
                                    LastNoteObjEventID[Lane] = AddEvent(tick, BmsEventCode.NoteObj + Lane, dt);
                                }
                            }

                            else if (ch < 0x50) // ch: 3x,4x
                            {
                                int Lane = ch - 0x30;

                                if (Lane == 0x08 || Lane == 0x09 || Lane == 0x18 || Lane == 0x19)
                                    infoData.is7Key = true;

                                AddEvent(tick, BmsEventCode.NoteHid + Lane, dt);
                            }
                            else if (ch < 0x70) // ch: 5x,6x
                            {
                                int Lane = ch - 0x50;

                                if (LastNoteOnEventID[Lane] == 0)
                                {
                                    LastNoteOnEventID[Lane] = AddEvent(tick, BmsEventCode.NoteOn + Lane, dt);
                                }
                                else
                                {
                                    eventData[LastNoteObjEventID[Lane]].Data2 = tick;
                                    AddEvent(tick, BmsEventCode.NoteOff + Lane, dt);
                                    LastNoteOnEventID[Lane] = 0;
                                }
                            }
                            else // 
                            {
                                switch (ch)
                                {
                                    case 0x99: AddEvent(tick, BmsEventCode.ChangeTextFromList, dt); break;
                                }
                            }
                        }
                    }
                }
                SortEventList();
                prePlay();
                return;
            }

            public void AddJudgeEvent(int PoorTick)
            {
                int events = this.events;
                BmsEvent[] prevEvent = new BmsEvent[32];

                for (int i = 0; i < events; i++)
                {
                    BmsEvent e = eventData[i];
                    if (e.EventType == BmsEventCode.TypeNote)
                    {
                        int Lane = e.EventLane;

                        int pos;

                        if (prevEvent[Lane].EventType == BmsEventCode.TypeNote)
                        {
                            pos = e.TickCount - Math.Min((e.TickCount - prevEvent[Lane].TickCount) / 2, PoorTick);

                            if (prevEvent[Lane].EventAttr == BmsEventCode.NoteObj)
                                AddEvent(Math.Min(pos, prevEvent[Lane].TickCount + PoorTick), BmsEventCode.JudgePoorNote + Lane, 0);
                            else if (prevEvent[Lane].EventAttr == BmsEventCode.NoteOn)
                                AddEvent(Math.Min(pos, prevEvent[Lane].TickCount + PoorTick), BmsEventCode.JudgePoorNoteOn + Lane, 0);
                            else if (prevEvent[Lane].EventAttr == BmsEventCode.NoteOff)
                                AddEvent(Math.Min(pos, prevEvent[Lane].TickCount + PoorTick), BmsEventCode.JudgePoorNoteOff + Lane, 0);
                        }
                        else
                        {
                            pos = Math.Max(e.TickCount - PoorTick, 0);
                        }

                        if (e.EventAttr == BmsEventCode.NoteObj)
                        {
                            AddEvent(pos, BmsEventCode.JudgeWaveAssign + Lane, e.Data);
                            AddEvent(pos, BmsEventCode.JudgeStartNote + Lane, e.TickCount);
                        }
                        else if (e.EventAttr == BmsEventCode.NoteOn)
                        {
                            AddEvent(pos, BmsEventCode.JudgeWaveAssign + Lane, e.Data);
                            AddEvent(pos, BmsEventCode.JudgeStartNoteOn + Lane, e.TickCount);
                        }
                        else if (e.EventAttr == BmsEventCode.NoteOff)
                        {
                            AddEvent(pos, BmsEventCode.JudgeStartNoteOff + Lane, e.TickCount);
                        }
                        prevEvent[Lane] = e;
                    }
                }

                // Ō̃m[g
                for (int i = 0; i < 32; i++)
                {
                    if (prevEvent[i].EventAttr == BmsEventCode.NoteObj)
                    {
                        AddEvent(prevEvent[i].TickCount + PoorTick, BmsEventCode.JudgePoorNote + i, 0);
                    }
                    else if (prevEvent[i].EventAttr == BmsEventCode.NoteOff)
                    {
                        AddEvent(prevEvent[i].TickCount + PoorTick, BmsEventCode.JudgeStartNoteOff + i, 0);
                    }
                }

                SortEventList();
                prePlay();
                return;
            }



            double CurrentBPM, CurrentMSPT, CurrentTPMS;

            void ChangeBpm(double newBpm)
            {
                CurrentBPM = newBpm;
                CurrentMSPT = 60000.0 / (CurrentBPM * (resolution / 4));
                CurrentTPMS = 1.0d / CurrentMSPT;
            }

            void prePlay()
            {
                ChangeBpm(infoData.bpm);

                double curTime = 0;
                int prevTick = 0;
                int notes = 0;

                for (int i = 0; i < events; i++)
                {
                    BmsEvent e = eventData[i];
                    switch (e.EventType)
                    {
                        case BmsEventCode.TypeSeqCtrl:
                            switch (e.Event)
                            {
                                case BmsEventCode.ChangeBPM: ChangeBpm(e.Data); break;
                                case BmsEventCode.ChangeBPMFromList: ChangeBpm(bpmlist[e.Data]); break;
                            }
                            break;
                        case BmsEventCode.TypeNote:
                            switch (e.EventAttr)
                            {
                                case BmsEventCode.NoteOn:
                                    notes++;
                                    break;

                                case BmsEventCode.NoteObj:
                                    notes++;
                                    break;
                            }
                            break;
                    }

                    curTime += (e.TickCount - prevTick) * CurrentMSPT;
                    prevTick = e.TickCount;
                }
                infoData.playtime = (int)curTime;
                infoData.totalnotes = notes;
                return;
            }

            /// <summary>
            /// Cxgǉ
            /// ǉ SortEventList() ĂŃ\[gB
            /// </summary>
            /// <param name="Tick">ʒu</param>
            /// <param name="Event">`l</param>
            /// <param name="Data">l</param>
            public int AddEvent(int Tick, int Event, int Data)
            {
                return AddEvent(Tick, Event, Data, 0);
            }

            public int AddEvent(int Tick, int Event, int Data, int Data2)
            {
                if (events >= MaxEvents)
                    return 0;
                eventData[events].Set(Tick, Event, Data, Data2);
                return events++;
            }

            /// <summary>
            /// CxgXg\[gB
            /// CxgǉIĂs邱ƁB
            /// </summary>
            /// 
            public void SortEventList()
            {
                BmsEvent[] work = new BmsEvent[events];
                for (int i = 0; ; i++)
                {
                    int n = 1 << i;
                    int m = n * 2;
                    for (int j = 0; ; j++)
                    {
                        int p = m * j;
                        int p1 = p, e1 = p1 + n;
                        int p2 = e1, e2 = (p2 + n > events) ? events : p2 + n;
                        int seg = e2 - p1;

                        if (seg <= n) break;

                        for (int k = 0; k < seg; k++)
                        {
                            if (p2 >= e2)
                                work[k] = eventData[p1++];
                            else if (p1 >= e1)
                                work[k] = eventData[p2++];
                            else if (eventData[p1].Event > eventData[p2].Event)
                                work[k] = eventData[p1++];
                            else
                                work[k] = eventData[p2++];
                        }
                        for (int k = 0; k < seg; k++)
                        {
                            eventData[k + p] = work[k];
                        }
                    }
                    if (n >= events) break;
                }

                for (int i = 0; ; i++)
                {
                    int n = 1 << i;
                    int m = n * 2;
                    for (int j = 0; ; j++)
                    {
                        int p = m * j;
                        int p1 = p, e1 = p1 + n;

                        int p2 = e1, e2 = Math.Min(p2 + n, events);
                        int seg = e2 - p1;

                        if (seg <= n) break;

                        for (int k = 0; k < seg; k++)
                        {
                            if (p2 >= e2)
                                work[k] = eventData[p1++];
                            else if (p1 >= e1)
                                work[k] = eventData[p2++];
                            else if (eventData[p1].TickCount < eventData[p2].TickCount)
                                work[k] = eventData[p1++];
                            else
                                work[k] = eventData[p2++];
                        }
                        for (int k = 0; k < seg; k++)
                        {
                            eventData[k + p] = work[k];
                        }
                    }
                    if (n >= events) break;
                }

            }

            static public int ParseInt(string str)
            {
                try
                {
                    return int.Parse(str);
                }
                catch (FormatException)
                {
                    return 0;
                }
            }

            static public double ParseDouble(string str)
            {
                try
                {
                    return double.Parse(str);
                }
                catch (FormatException)
                {
                    return 0.0f;
                }
            }

            static public int ParseInt36(string str)
            {
                int value = 0;
                foreach (char c in str)
                {
                    if (c <= ' ') { continue; }
                    if (('0' <= c) && (c <= '9')) { value = value * 36 + c - '0'; continue; }
                    if (('A' <= c) && (c <= 'Z')) { value = value * 36 + c - 'A' + 10; continue; }
                    if (('a' <= c) && (c <= 'z')) { value = value * 36 + c - 'a' + 10; continue; }
                    return -1;
                }
                return value;
            }

            static public int ParseInt16(string str)
            {
                try
                {
                    return int.Parse(str, System.Globalization.NumberStyles.HexNumber);
                }
                catch (FormatException)
                {
                    return -1;
                }
            }

        }
    }
}